from packaging.version import InvalidVersion, Version

from .. import NotSet
from ..errors import InvalidVersionError, _IgnoredVersionError
from .distro import Distro
from .hab_base import HabBase
from .meta import hab_property


class DistroVersion(HabBase):
    """A specific version of the loaded `Distro`'s. Including its requirements,
    aliases and environment variables."""

    _context_method = "name"
    _placeholder = Distro

    def __init__(self, *args, **kwargs):
        self._alias_mods = NotSet
        super().__init__(*args, **kwargs)

    def _cache(self):
        return self.resolver.site.cache.distro_paths(flat=True)

    def _resolve_version(self, data, filename):
        """Sets and returns self.version to the correct value for this distro.

        This resolves the version from the several ways it can be stored to
        simplify deployment and development. See InvalidVersionError for details
        on how distros version can be set.

        Raises:
            _IgnoredVersionError: Internal use, this version should not be processed.
            InvalidVersionError: Raised if the version could not be resolved.
        """
        version_txt = self.dirname / ".hab_version.txt"

        if "version" in data:
            self.version = data["version"]
            return self.version
        elif version_txt.exists():
            self.version = version_txt.open().read().strip()
            return self.version

        # If version is not defined in json data extract it from the parent
        # directory name. This allows for simpler distribution without needing
        # to modify version controlled files.
        try:
            self.version = self.dirname.name
            return self.version
        except InvalidVersion:
            """The parent directory was not a valid version, attempt to get a
            version using setuptools_scm.
            """
            try:
                from setuptools_scm import get_version
            except ImportError as error:
                raise InvalidVersionError(filename, error=error) from None

            def check_ignored_version():
                if self.dirname.name in self.resolver.ignored:
                    # This object is not added to the forest until super is called
                    raise _IgnoredVersionError(
                        'Skipping "{}" its dirname is in the ignored list.'.format(
                            filename
                        )
                    ) from None

            try:
                self.version = get_version(
                    root=self.dirname, version_scheme="release-branch-semver"
                )
                return self.version
            except LookupError:
                check_ignored_version()
                raise InvalidVersionError(filename) from None
            except Exception as error:
                check_ignored_version()
                # To make debugging easier include the original exception
                raise InvalidVersionError(filename, error=error) from None

    @hab_property()
    def aliases(self):
        """List of the names and commands that need created to launch desired
        applications."""
        return self.frozen_data.get("aliases", NotSet)

    @aliases.setter
    def aliases(self, aliases):
        self.frozen_data["aliases"] = aliases

    # Note: 'alias_mods' needs to be processed before 'environment'
    @hab_property(process_order=50)
    def alias_mods(self):
        """Dict of modifications that need to be made on aliases.
        These are used to modify the original configuration of an alias by another
        distro or config. This allows a plugin to add an environment variable to
        a specific alias even though the alias is defined by another distro/config.
        """
        return self._alias_mods

    def _load(self, filename, cached=True):
        """Sets self.filename and parses the json file returning the data."""
        ret = super()._load(filename, cached=cached)

        # Resolve the version from the various supported ways its stored.
        self._resolve_version(ret, filename)

        if not cached:
            # Ensure the version is stored on the returned dictionary
            ret["version"] = str(self.version)
        return ret

    def load(self, filename):
        # Fill in the DistroVersion specific settings before calling super
        data = self._load(filename)

        self.aliases = data.get("aliases", NotSet)
        # Store any alias_mods, they will be processed later when flattening
        self._alias_mods = data.get("alias_mods", NotSet)

        # The name should be the version == specifier.
        self.distro_name = data.get("name")
        self.name = "{}=={}".format(self.distro_name, self.version)

        data = super().load(filename, data=data)

        return data

    @hab_property()
    def version(self):
        return super().version

    @version.setter
    def version(self, version):
        # NOTE: super doesn't work for a @property.setter
        if version and not isinstance(version, Version):
            version = Version(version)
        self.frozen_data["version"] = version



import logging
from copy import deepcopy

from .. import NotSet
from .config import Config
from .meta import hab_property

logger = logging.getLogger(__name__)


class UnfrozenConfig(Config):
    def __init__(self, frozen_data, resolver, uri=NotSet, forest=None):
        super().__init__(None, resolver)
        self.frozen_data = deepcopy(frozen_data)

        # Restore the HAB_URI env variable that was removed during freeze
        for platform in self.frozen_data.setdefault("environment", {}):
            self.frozen_data["environment"][platform].setdefault(
                "HAB_URI", self.frozen_data["uri"]
            )

    @classmethod
    def _dump_versions(cls, value, verbosity=0, color=None):
        """Returns the version information for this object as a list of strings."""
        # Frozen versions are already a list of strings
        return sorted(value)

    # Note: 'alias_mods' needs to be processed before 'environment'
    @hab_property(verbosity=None, process_order=50)
    def alias_mods(self):
        """Returns NotSet. Any alias_mods have already been baked into aliases."""
        return NotSet

    @hab_property(verbosity=2)
    def inherits(self):
        """Returns False. Unfrozen configurations can not inherit from other configs."""
        return False

    @property
    def fullpath(self):
        return self.frozen_data["uri"]

    @property
    def uri(self):
        return self.frozen_data["uri"]

    @hab_property(verbosity=1)
    def versions(self):
        return self.frozen_data.get("versions", [])



import logging
from copy import deepcopy

from .. import NotSet, utils
from ..merge_dict import MergeDict
from .config import Config
from .meta import hab_property

logger = logging.getLogger(__name__)


class FlatConfig(Config):
    """A fully resolved and flattened Config object. Any values NotSet on the Config
    this is built from, are attempted to be set from the parents of that Config. If
    still not found, it will attempt to find the value from a matching config on the
    Default tree instead."""

    def __init__(self, original_node, resolver, uri=NotSet):
        super().__init__(original_node.forest, resolver)
        self.original_node = original_node
        self.filename = original_node.filename
        self.frozen_data["context"] = original_node.context
        self._uri = uri
        # Caches a copy of modifications that eventually need applied to aliases
        self._alias_mods = NotSet
        # Copy the properties from the inheritance system
        self._collect_values(self.original_node)
        self._finalize_values()

    def _finalize_values(self):
        """Post processing done after `_collect_values` is run."""

        # `_process_version` needs the global environment variables populated
        # so populating per-alias env var's properly inherit the global variables.
        # This call ensures that `self.frozen_data["environment"]` is populated.
        self.environment

        # Run any configured entry_points before aliases are calculated
        self.resolver.site.run_entry_points_for_group("hab.cfg.reduce.env", cfg=self)

        # Ensure distros are populated before versions get processed
        if self.distros is NotSet:
            if self.resolver.forced_requirements:
                self.distros = self.resolver.forced_requirements
            else:
                self.distros = {}

        # Process version aliases, merging global env vars.
        platform_aliases = {}
        self.frozen_data["aliases"] = platform_aliases
        for version in self.versions:
            for platform, alias, data in self._process_version(
                version, existing=platform_aliases
            ):
                platform_aliases.setdefault(platform, {})[alias] = data

        # Run any configured entry_points before finishing
        self.resolver.site.run_entry_points_for_group(
            "hab.cfg.reduce.finalize", cfg=self
        )

    def _process_version(self, version, existing=None):
        """Generator that yields each finalized alias definition dictionary
        to be stored in the frozen_data for aliases.

        Args:
            version: A Version to process.
            existing: A record of the aliases already processed. This is used
                as an early out when encountering a duplicate alias name.
                Duplicate aliases are allowed, but only uses the first encountered.

        Yields:
            tuple: 3 item tuple containing (platform, alias, alias_spec). The
                alias_spec is stored in frozen_data under the other two keys.
        """
        if not version.aliases:
            # There are no aliases to process, so we can simply exit
            return

        # TODO: Add support for the '*'' platform
        for platform in self.resolver.site["platforms"]:
            aliases_def = version.aliases.get(platform, [])
            aliases = [a[1] for a in aliases_def]

            merger = MergeDict(platforms=[platform], relative_root=version.dirname)

            for i, alias in enumerate(aliases_def):
                alias_name = alias[0]

                # Configure the merger for this version
                merger.formatter = version.format_environment_value
                merger.validator = self.check_environment

                # Only process an alias the first time it is encountered
                if existing and alias_name in existing.get(platform, {}):
                    logger.info(
                        f'Skipping duplicate alias "{alias_name}" for '
                        f'"{version.name}" on "{platform}"'
                    )
                    continue

                host_platform = utils.Platform.name()
                # Ensure that we always have a dictionary for aliases
                data = version.format_environment_value(aliases[i])
                if not isinstance(data, dict):
                    data = dict(cmd=data)

                mods = self._alias_mods.get(alias_name, [])
                if "environment" not in data and not mods:
                    yield platform, alias_name, data
                    continue

                # Merge and flatten the alias and global env var for the platform

                # If a per-alias env var is also defined in the global env var's
                # managed by hab, use that as the base env var, otherwise overwrite
                # non-hab managed env vars as we do for global env vars.
                global_env = self.frozen_data["environment"].get(host_platform, {})
                merged_env = {}

                def extract_global_keys(operation, merged_env, global_env):
                    for plat in operation.values():
                        for key in plat:
                            if key not in merged_env and key in global_env:
                                merged_env[key] = global_env[key]

                environment = data.get("environment", {})
                extract_global_keys(environment, merged_env, global_env)
                for mod in mods:
                    env = mod.get("environment", {}).get(host_platform, {})
                    extract_global_keys(env, merged_env, global_env)

                merged_env = {platform: merged_env}

                # Update the global env vars with per-alias env vars
                merger.apply_platform_wildcards(environment, output=merged_env)
                # Apply any alias_mods specified for this alias.
                for mod in mods:
                    env = mod.get("environment")
                    if env:
                        merger.apply_platform_wildcards(env, output=merged_env)

                # Update the alias environment removing the redundant platform spec
                data["environment"] = merged_env[platform]
                yield platform, alias_name, data

    # Note: 'alias_mods' needs to be processed before 'environment'
    @hab_property(verbosity=None, process_order=50)
    def alias_mods(self):
        """Override the hab_property decorator's verbosity to hide this from the
        dump. It doesn't make sense to show this property on FlatConfig, but it's
        important to show it on the super Config class for debugging.
        """
        return super().alias_mods

    # Note: 'alias_mods' and 'distros' needs to be processed before 'environment'
    @hab_property(verbosity=2, process_order=80)
    def environment(self):
        """A resolved set of environment variables for this platform that should
        be applied to configure an environment. Any values set to None indicate
        that the variable should be unset.
        """
        if self.frozen_data.get("environment") is None:
            super().environment
            # Add any environment variables defined by the linked versions
            for version in self.versions:
                self.merge_environment(version.environment_config, obj=version)
            # Add the HAB_URI env var for each platform so scripts know they are
            # in an activated hab environment and the original uri the user requested.
            for platform in self.resolver.site["platforms"]:
                self.frozen_data.setdefault("environment", {}).setdefault(platform, {})[
                    "HAB_URI"
                ] = [self.uri]

        return self.frozen_data["environment"].get(utils.Platform.name(), {})

    @property
    def fullpath(self):
        if self.context:
            return self.separator.join([name for name in self.context] + [self.name])
        return self.name

    def freeze(self):
        """Returns information that can be used to create a unfrozen copy of
        this configuration.
        """

        # ensure the version environments are flattened into the environment
        self.environment

        frozen_data = deepcopy(self.frozen_data)
        frozen_data["uri"] = self.uri
        if "versions" in self.frozen_data:
            frozen_data["versions"] = [v.name for v in self.frozen_data["versions"]]

        # Simplify the output data by removing un-needed and duplicated items
        for platform in frozen_data.get("environment", {}):
            frozen_data["environment"][platform].pop("HAB_URI", None)

        # No need to store the environment_config in a freeze
        frozen_data.pop("environment_config", None)

        # No need to store this it will always be False
        frozen_data.pop("inherits", None)

        # Remove any empty properties that are not required
        for key in ("aliases", "versions"):
            if key in frozen_data and not frozen_data[key]:
                frozen_data.pop(key, None)

        return frozen_data

    @hab_property(verbosity=1)
    def versions(self):
        distros = self.distros
        if distros is NotSet:
            return []
        if distros == []:
            distros = {}

        # Lazily load the contents of versions the first time it's called
        if "versions" not in self.frozen_data:
            versions = []
            if self._alias_mods is NotSet:
                self._alias_mods = {}
            self.frozen_data["versions"] = versions

            reqs = self.resolver.resolve_requirements(distros)
            for req in reqs.values():
                version = self.resolver.find_distro(req)
                versions.append(version)

                # If this version defines any alias_mods, store them for later
                if version.alias_mods:
                    # Format the alias environment at this point so any path
                    # based variables like {relative_root} are resolved against
                    # the version's directory not the alias being modified
                    mods = version.format_environment_value(version.alias_mods)
                    for name, mod in mods.items():
                        self._alias_mods.setdefault(name, []).append(mod)

        return self.frozen_data["versions"]



import datetime
import json
import logging

import pytest

from hab import user_prefs, utils


def test_user_prefs_filename():
    """Check that user_prefs_filename generates the expected file paths."""
    # Check the default for filename
    path = utils.Platform.user_prefs_filename()
    assert path.name == ".hab_user_prefs.json"

    # Check overriding filename
    new = utils.Platform.user_prefs_filename(filename="test.json")
    assert new == path.parent / "test.json"


def test_configure_logging(monkeypatch, tmpdir):
    # Ensure we can read/write logging prefs, but using the test dir.
    monkeypatch.setenv("HOME", str(tmpdir))
    monkeypatch.setenv("LOCALAPPDATA", str(tmpdir))

    logger = logging.getLogger("hab.test")
    default = utils.Platform.user_prefs_filename(".hab_logging_prefs.json")
    custom = tmpdir / "test.json"

    logging_cfg = {
        "version": 1,
        "incremental": True,
        "loggers": {"hab.test": {"level": 10}},
    }

    # The default file doesn't exist yet, no configuration is loaded
    assert not utils.Platform.configure_logging()
    # The logger's level is not configured yet
    assert logger.level == 0

    # Create the configuration and ensure it gets loaded
    with default.open("w") as fle:
        json.dump(logging_cfg, fle)
    assert utils.Platform.configure_logging()
    # The logger had its level set to 10 by the config
    assert logger.level == 10

    # Check that passing a filename is respected
    logging_cfg["loggers"]["hab.test"]["level"] = 30
    with custom.open("w") as fle:
        json.dump(logging_cfg, fle)
    assert utils.Platform.configure_logging(filename=custom)
    # The logger had its level set to 10 by the config
    assert logger.level == 30


def test_filename(uncached_resolver, tmpdir):
    prefs = uncached_resolver.user_prefs(load=True)
    # Defaults to Platform path
    assert prefs.filename == utils.Platform.user_prefs_filename()

    filename = tmpdir / ".hab_user_prefs.json"
    prefs.filename = filename
    assert prefs.filename == filename

    # If the file doesn't exist on disk, load returns False
    assert prefs.load() is False


def test_enabled_no_default(uncached_resolver):
    """If prefs_default is not specified default to disabled."""
    uncached_resolver.site.pop("prefs_default", None)
    prefs = user_prefs.UserPrefs(uncached_resolver)
    assert prefs.enabled is False


@pytest.mark.parametrize(
    "setting,default,value,check",
    (
        # If prefs_default is disabled, the result is False for all values
        (["disabled"], False, None, False),
        (["disabled"], False, True, False),
        (["disabled"], False, False, False),
        # Enabling prefs by default
        (["--prefs"], True, None, True),
        (["--prefs"], True, True, True),
        (["--prefs"], True, False, False),
        ([True], True, True, True),
        # Disabling prefs by default
        (["--no-prefs"], False, None, False),
        (["--no-prefs"], False, True, True),
        (["--no-prefs"], False, False, False),
        ([False], False, False, False),
    ),
)
def test_enabled(uncached_resolver, setting, default, value, check):
    uncached_resolver.site["prefs_default"] = setting
    prefs = user_prefs.UserPrefs(uncached_resolver)

    assert prefs.enabled == default
    prefs.enabled = value
    assert prefs.enabled == check


def test_timeout(uncached_resolver):
    uncached_resolver.site.pop("prefs_uri_timeout", None)
    prefs = user_prefs.UserPrefs(uncached_resolver)

    def set_uri_last_changed(**kwargs):
        d = datetime.datetime.today()
        d = d - datetime.timedelta(**kwargs)
        prefs["uri_last_changed"] = d.isoformat()

    # uri_is_timedout works if last_chanded is not set
    prefs.pop("uri_last_changed", None)
    assert prefs.uri_is_timedout is False

    # prefs_uri_timeout is not set
    set_uri_last_changed(minutes=3)
    assert prefs.uri_timeout is None

    # prefs_uri_timeout enables uri timeout
    uncached_resolver.site["prefs_uri_timeout"] = dict(minutes=5)
    assert prefs.uri_timeout == datetime.timedelta(minutes=5)
    assert prefs.uri_is_timedout is False
    set_uri_last_changed(minutes=6)
    assert prefs.uri_is_timedout is True

    uncached_resolver.site["prefs_uri_timeout"] = dict(days=30)
    assert prefs.uri_timeout == datetime.timedelta(days=30)
    assert prefs.uri_is_timedout is False
    set_uri_last_changed(days=40)
    assert prefs.uri_is_timedout is True


def test_uri(uncached_resolver, tmpdir, monkeypatch):
    uncached_resolver.site.pop("prefs_uri_timeout", None)

    # Force the prefs to be saved into the test directory.
    if utils.Platform.name() == "windows":
        monkeypatch.setenv("LOCALAPPDATA", str(tmpdir))
    else:
        monkeypatch.setenv("HOME", str(tmpdir))
    # Ensure we are using the modified filepath
    prefs_a = user_prefs.UserPrefs(uncached_resolver)
    assert prefs_a.filename.parent == tmpdir

    # No preferences are saved
    prefs_b = user_prefs.UserPrefs(uncached_resolver)
    prefs_b._enabled = False
    assert prefs_b.uri is None

    # Preferences store an uri
    prefs_b["uri"] = "app/aliased"
    prefs_b.save()
    prefs_c = user_prefs.UserPrefs(uncached_resolver)
    # Prefs are disabled
    prefs_c._enabled = False
    assert prefs_c.uri is None
    # Prefs are enabled
    prefs_c._enabled = True
    assert prefs_c.uri == "app/aliased"

    # Preferences store timeout information
    last = datetime.datetime.today()
    last = last - datetime.timedelta(hours=1)
    prefs_c["uri_last_changed"] = last.isoformat()
    prefs_c.save()

    prefs_d = user_prefs.UserPrefs(uncached_resolver)
    prefs_d._enabled = True
    # Timeout has not expired
    uncached_resolver.site["prefs_uri_timeout"] = dict(hours=2)
    assert prefs_d.uri_check().timedout is False
    # Timeout has expired
    uncached_resolver.site["prefs_uri_timeout"] = dict(minutes=5)
    assert prefs_d.uri_check().timedout is True

    # Check that uri.setter is processed correctly
    prefs_e = user_prefs.UserPrefs(uncached_resolver)

    # uri.setter is ignored if prefs are disabled
    prefs_e._enabled = False
    prefs_e.uri = "app/aliased/mod"
    assert "uri" not in prefs_e
    assert prefs_e.uri is None

    # uri.setter updates the uri and sets the timestamp
    assert "uri_last_changed" not in prefs_e
    prefs_e._enabled = True
    assert "uri" not in prefs_e
    prefs_e.uri = "app/aliased/mod"
    assert prefs_e["uri"] == "app/aliased/mod"
    assert prefs_e.uri == "app/aliased/mod"
    # The timeout was properly stored
    assert prefs_e["uri_last_changed"].date() == datetime.date.today()

    # Check that the file was actually written
    assert prefs_e.filename.exists()
    assert prefs_e.filename == utils.Platform.user_prefs_filename()

    # Check if UriObj.__str__() is passing the uri contents
    prefs_g = user_prefs.UriObj("test/uri")
    assert prefs_g.__str__() == "test/uri"

    # Check UriObj.__str__ returns a string even if uri is None
    prefs_g = user_prefs.UriObj()
    assert isinstance(prefs_g.__str__(), str)


@pytest.mark.parametrize(
    "test_text",
    (
        # File was opened in write mode but no contents written
        None,
        # Process was killed half way through writing json output.
        '{\n    "uri": "app',
    ),
)
def test_corruption(uncached_resolver, tmpdir, monkeypatch, caplog, test_text):
    """Check how UserPrefs handles trying to load an incomplete or empty existing
    json document.
    """
    caplog.set_level(logging.INFO)

    def assert_log_exists(level, msg):
        for record in caplog.records:
            if record.levelname != level:
                continue
            if msg in record.message:
                print(record)
                break
        else:
            raise AssertionError(
                f"No logging message was made matching: {level} and {msg}"
            )

    # TODO: This duplicated code should probably be moved into a fixture
    # Force the prefs to be saved into the test directory.
    if utils.Platform.name() == "windows":
        monkeypatch.setenv("LOCALAPPDATA", str(tmpdir))
    else:
        monkeypatch.setenv("HOME", str(tmpdir))
    # Ensure we are using the modified filepath
    prefs = user_prefs.UserPrefs(uncached_resolver)
    assert prefs.filename.parent == tmpdir

    prefs_file = tmpdir / ".hab_user_prefs.json"

    # Create a empty file that won't resolve properly for json.
    with prefs_file.open("w") as fle:
        if test_text:
            fle.write(test_text)

    # Check that the expected log messages are emitted when invalid
    # json file contents are encountered
    caplog.clear()
    prefs = user_prefs.UserPrefs(uncached_resolver)
    # Even with invalid contents True will be returned
    assert prefs.load()
    # When corrupt prefs are encountered, default empty dict results
    assert prefs == {}

    assert_log_exists("WARNING", "User pref file corrupt, resetting.")
    assert_log_exists("INFO", "User pref exception suppressed:")



import os

import pytest

from hab import utils
from hab.formatter import Formatter
from hab.parsers import Config


@pytest.mark.parametrize(
    "language,shell,pathsep",
    (
        ("sh", "-$PATH-", "-:-"),
        # Bash formatting is different on windows for env vars
        ("shwin", "-$PATH-", "-:-"),
        ("ps", "-$env:PATH-", "-;-"),
        ("batch", "-%PATH%-", "-;-"),
        (None, "-{PATH!e}-", "-{;}-"),
    ),
)
def test_e_format(language, shell, pathsep):
    """Check that "{VAR_NAME!e}" is properly formatted."""
    path = os.environ["PATH"]

    # Check that "!e" is converted to the correct shell specific specifier.
    assert Formatter(language).format("-{PATH!e}-") == shell

    # Check that "!e" uses the env var value if `expand=True` not the shell specifier.
    assert Formatter(language, expand=True).format("-{PATH!e}-") == f"-{path}-"

    # Check that the pathsep variable `{;}` is converted to the correct value
    assert Formatter(language).format("-{;}-") == pathsep


def test_language_from_ext(monkeypatch):
    # Arbitrary values are not modified
    assert Formatter.language_from_ext(".abc") == ".abc"
    assert Formatter.language_from_ext("anything") == "anything"

    # Test that known file exts are translated to the shell name
    assert Formatter.language_from_ext(".bat") == "batch"
    assert Formatter.language_from_ext(".cmd") == "batch"
    assert Formatter.language_from_ext(".ps1") == "ps"

    # Bash formatting is different on windows for env vars
    with monkeypatch.context() as m:
        m.setattr(utils, "Platform", utils.WinPlatform)
        assert Formatter.language_from_ext(".sh") == "shwin"
        assert Formatter.language_from_ext("") == "shwin"

        m.setattr(utils, "Platform", utils.LinuxPlatform)
        assert Formatter.language_from_ext(".sh") == "sh"
        assert Formatter.language_from_ext("") == "sh"


def test_format_environment_value(uncached_resolver):
    forest = {}
    config = Config(forest, uncached_resolver)

    # test_format_environment_value doesn't replace the special formatters.
    # This allows us to delay these formats to only when creating the final
    # shell scripts, not the first time we evaluate environment variables.
    value = "a{;}b;c:{PATH!e}{;}d"
    assert config.format_environment_value(value, ext=None) == value



import glob
import json
import logging
from pathlib import Path

from packaging.version import InvalidVersion

from . import utils
from .errors import InvalidVersionError, _IgnoredVersionError

logger = logging.getLogger(__name__)


class Cache:
    """Used to save/restore cached data to speed up initialization of hab.

    The caches are stored per-site file as file next to the site file using the
    same stem name. (Ie by default studio.json would have a cache file called
    studio.habcache).

    If this cache file exists it is used unless enabled is set to False. Cache
    files are useful when you have some sort of CI setup to ensure the cache is
    re-generated using `save_cache` any time you make changes to configs or
    distros that site file references.

    Properties:
        cache_template (dict): The str.format template used to find the cache files.
            This template requires the kwarg `stem`. This template can be overridden
            by the first `site_cache_file_template` property in site.
        enabled (bool): Used to disable using of the cached data forcing a full
            glob and parse of files described by all site files.
    """

    supported_version = 1

    def __init__(self, site):
        self.site = site
        self._cache = None
        self.enabled = True

        # Get the template filename used to find the cache files on disk
        self.cache_template = self.site["site_cache_file_template"][0]

    @property
    def cached_keys(self):
        """A dict of cache keys and how they should be processed.
        {Name of key to cache: ("relative file glob", class used to process)}
        """
        try:
            return self._cached_keys
        except AttributeError:
            pass

        from .parsers import Config, DistroVersion

        self._cached_keys = {
            "config_paths": ("*.json", Config),
            "distro_paths": ("*/.hab.json", DistroVersion),
        }
        return self._cached_keys

    def cache(self, force=False):
        if not self.enabled:
            # If caching is disabled, never attempt to load the cache
            return {}

        if self._cache is not None and not force:
            return self._cache

        self._cache = {}

        # Process caches from right to left. This makes it so the left most
        # cache_file is respected if any paths are duplicated.
        for path in reversed(self.site.paths):
            cache_file = self.site_cache_path(path)
            if cache_file.is_file():
                logger.debug(f"Site cache loading: {cache_file!s}")
                self.load_cache(cache_file)

        # Create a flattened cache removing the glob paths.
        flat_cache = {key: {} for key in self.cached_keys}
        for key in self._cache:
            for values in self._cache.get(key, {}).values():
                flat_cache[key].update(values)

        self._cache["flat"] = flat_cache

        return self._cache

    def clear(self):
        """Reset the cache forcing it to reload the next time its used."""
        if self._cache:
            logger.debug("Site cache contents cleared")
        self._cache = None

    def config_paths(self, flat=False):
        if flat:
            return self.cache().get("flat", {}).get("config_paths", {})
        return self.cache().get("config_paths", {})

    def distro_paths(self, flat=False):
        if flat:
            return self.cache().get("flat", {}).get("distro_paths", {})
        return self.cache().get("distro_paths", {})

    def generate_cache(self, resolver, site_file, version=1):
        """Generate a cache file of the current state defined by this site file.
        This contains the raw values of each URI config and distro file including
        version. If this cache exists it is used instead of searching the file
        system for each path defined in config_paths or distro_paths defined in
        the provided site file. Use this method any time changes are made that
        hab needs to be aware of. Caching is enabled by the existence of this file.
        """
        from .site import Site

        # Indicate the version specification this habcache file conforms to.
        output = {"version": version}

        # read the site file to get paths to process
        temp_site = Site([site_file])
        # Use this to convert platform specific paths to generic variables
        platform_path_key = resolver.site.platform_path_key

        for key, stats in self.cached_keys.items():
            glob_str, cls = stats
            # Process each glob dir defined for this site
            for dirname in temp_site.get(key, []):
                cfg_paths = output.setdefault(key, {}).setdefault(
                    platform_path_key(dirname).as_posix(), {}
                )

                # Add each found hab config to the cache
                for path in sorted(glob.glob(str(dirname / glob_str))):
                    path = Path(path)
                    try:
                        data = cls(forest={}, resolver=resolver)._load(
                            path, cached=False
                        )
                    except (
                        InvalidVersion,
                        InvalidVersionError,
                        _IgnoredVersionError,
                    ) as error:
                        logger.debug(str(error))
                    else:
                        cfg_paths[platform_path_key(path).as_posix()] = data

        return output

    @classmethod
    def iter_cache_paths(cls, name, paths, cache, glob_str=None, include_path=True):
        """Yields path information stored in the cache falling back to glob if
        not cached.

        Yields:
            dirname: Each path stored in paths.
            path
        """
        for dirname in paths:
            dn_posix = dirname.as_posix()
            cached = dn_posix in cache
            if cached:
                logger.debug(f"Using cache for {name} dir: {dn_posix}")
                paths = cache[dn_posix]
            else:
                logger.debug(f"Using glob for {name} dir: {dirname}")
                # Fallback to globing the file system
                if glob_str:
                    paths = sorted(glob.glob(str(dirname / glob_str)))
                else:
                    paths = []
            if not include_path:
                yield dirname, None, cached
            else:
                for path in paths:
                    yield dirname, path, cached

    def load_cache(self, filename, platform=None):
        """For each glob dir add or replace the contents. If a previous cache
        has the same glob dir, it's cache is ignored. This expects that
        load_cache is called from right to left for each path in `self.site.path`.
        """

        def cache_to_platform(cache, mappings):
            """Restore the cross-platform variables to current platform paths."""
            ret = {}
            for glob_str, files in cache.items():
                new_glob = glob_str.format(**mappings)
                new_glob = Path(new_glob).as_posix()
                new_files = {}
                for key, value in files.items():
                    new_key = key.format(**mappings)
                    new_key = Path(new_key).as_posix()
                    new_files[new_key] = value
                ret[new_glob] = new_files

            return ret

        if platform is None:
            platform = utils.Platform.name()

        contents = utils.load_json_file(filename)

        # If the cache was saved by a newer habcache system, warn that we are
        # unable to load the cache but don't raise an exception
        if contents["version"] > self.supported_version:
            logger.warning(
                f"File is using a unsupported habcache version {contents['version']}. "
                f"Only versions > {self.supported_version} are supported, ignoring {filename}"
            )
            return

        mappings = self.site.get("platform_path_maps", {})
        mappings = {key: value[platform] for key, value in mappings.items()}
        for key in self.cached_keys:
            if key in contents:
                cache = cache_to_platform(contents[key], mappings)
                self._cache.setdefault(key, {}).update(cache)

    def save_cache(self, resolver, site_file, version=1):
        cache_file = self.site_cache_path(site_file)
        cache = self.generate_cache(resolver, site_file, version=version)

        with cache_file.open("w") as fle:
            json.dump(cache, fle, indent=4, cls=utils.HabJsonEncoder)
        return cache_file

    def site_cache_path(self, path):
        """Returns the name of the cache file for the given site file."""
        return path.parent / self.cache_template.format(stem=path.stem)



from .config import Config
from .distro import Distro
from .distro_version import DistroVersion
from .flat_config import FlatConfig
from .hab_base import HabBase
from .meta import HabMeta, hab_property
from .placeholder import Placeholder
from .unfrozen_config import UnfrozenConfig

__all__ = [
    "Config",
    "Distro",
    "DistroVersion",
    "FlatConfig",
    "hab_property",
    "HabBase",
    "HabMeta",
    "Placeholder",
    "UnfrozenConfig",
]



class HabError(Exception):
    """Base class for all hab errors."""


class DuplicateJsonError(HabError):
    """Raised if the same context/name is re-defined inside a single config/distro
    path definition. This is allowed for unique `Resolver.config_paths` and the first
    path that defines it is the one used, but if the same context is used inside a
    single path this error is raised.
    """


class RequirementError(HabError):
    """Base class for Requirement parsing errors."""


class _IgnoredVersionError(RequirementError):
    """Internal exception raised if a distro version is ignored."""


class MaxRedirectError(RequirementError):
    """The maximum number of redirects was reached without resolving successfully."""


class InvalidRequirementError(RequirementError):
    """Raised if unable to resolve a given requirement."""


class InvalidVersionError(LookupError):
    """Provides info on resolving why it was unable to generate a valid version number"""

    default_message = (
        'Hab was unable to determine the version for "{filename}".\n'
        "  The version is defined in one of several ways checked in this order:\n"
        "  1. The version property in `.hab.json`.\n"
        "  2. A `.hab_version.txt` file next to `.hab.json`.\n"
        "  3. `.hab.json`'s parent directory name.\n"
        "  4. setuptools_scm can get a version from version control.\n"
        "  The preferred method is #3 for deployed releases. #4 is the "
        "preferred method for developer working copies."
    )

    def __init__(self, filename, error=None, message=None):
        self.filename = filename
        self.error = error
        if message is None:
            message = self.default_message
        self.message = message

    def __str__(self):
        ret = self.message.format(filename=self.filename)
        if self.error:
            ret = f"[{type(self.error).__name__}] {self.error}\n{ret}"
        return ret


class ReservedVariableNameError(HabError):
    """Raised if a custom variable uses a reserved variable name."""

    _reserved_variable_names = {"relative_root", ";"}
    """A set of variable names hab reserved for hab use and should not be defined
    by custom variables."""

    def __init__(self, invalid, filename):
        self.filename = filename
        msg = (
            f"{', '.join(sorted(invalid))!r} are reserved variable name(s) for "
            "hab and can not be used in the variables section. "
            f"Filename: '{self.filename}'"
        )
        super().__init__(msg)



import json
from pathlib import Path

import pytest
from jinja2 import Environment, FileSystemLoader

from hab import errors, utils

# Create a test for each dir in reference_scripts. We pass the name of the folder
# so test selection matches the name of the folder instead of something abstract
# like [reference5].
reference_scripts = Path(__file__).parent / "reference_scripts"
reference_names = [x.name for x in reference_scripts.iterdir()]


@pytest.mark.parametrize("reference_name", reference_names)
def test_scripts(uncached_resolver, tmpdir, monkeypatch, config_root, reference_name):
    """Checks all of the scripts HabBase.write_script generates for a specified
    set of arguments.

    This is a parametrized test, and each folder inside the tests/reference_scripts
    directory adds a test. Each directory needs a spec.json file similar to this::

        {
            "description": "[shell and platform]: Tests `hab [command] [uri]`",
            "ext": ".sh",
            "launch": null,
            "exiting": false,
            "args": null,
            "create_launch": false,
            "platform": "linux",
            "uri": "not_set/child"
        }

    A reference directory needs to have at least one script file generated by hab,
    and all files are expected to be created by write_script.

    To add a new test, create the new directory and add the desired spec file.
    Run the test(it will fail) and in the pytest temp directory you can find the
    script files that hab generated for your spec. Copy them into your reference
    directory. You will then need to replace the hard coded file paths with jinja2
    templates to ensure your test will pass on the next run of pytest, as well
    as when the git checkout is in a different location. There are a few required
    replacements needed:

    1. Replace the path to the test's tempdir with `{{ tmpdir }}`. If joining
        paths to tmpdir, do it as part of the format to ensure correct slash
        direction. `{{ tmpdir / "filename.txt" }}`
    2. Replace the path to your checkout's tests directory with `{{ config_root }}`
        making sure to convert back slashes to forward slashes
    3. Replace the HAB_FREEZE value with `{{ freeze }}`.
    """
    reference = reference_scripts / reference_name
    spec = json.load((reference / "spec.json").open())

    if "description" in spec:
        print(f"Testing: {spec['description']}")

    # Script formatting is subtly different on different platforms, ensure we
    # are testing the requested platform, not the current one.
    platform = spec["platform"]
    assert platform in ("linux", "osx", "win32")
    monkeypatch.setattr(utils, "Platform", utils.BasePlatform.get_platform(platform))

    cfg = uncached_resolver.resolve(spec["uri"])
    reference = config_root / "reference_scripts" / reference_name
    ext = spec["ext"]

    cfg.write_script(
        str(tmpdir),
        ext=ext,
        launch=spec["launch"],
        exit=spec["exiting"],
        args=spec["args"],
        create_launch=spec["create_launch"],
    )

    # Generate the correct testing path for the shell this test is targeting
    func = utils.path_forward_slash
    if platform == "win32" and ext == ".sh":
        func = utils.cygpath

    # Test specific values to update the reference templates with for comparison
    replacements = {
        "freeze": utils.encode_freeze(cfg.freeze()),
        # Replace the path to the test's tempdir with `{{ tmpdir }}`
        "tmpdir": tmpdir,
        # Replace the path to your checkout's tests directory with
        # `{{ config_root }}` making sure to convert back slashes to forward slashes
        "config_root": func(config_root),
        # For bash on windows, we are not converting "cmd" to cygpaths as it's
        # not required and we would then need to account for multiple paths in
        # the same string.
        "config_root_alias": utils.path_forward_slash(config_root),
    }

    def check_file(item):
        rel_path = item.relative_to(reference)
        generated = tmpdir / rel_path

        # Fill in the dynamic variables in the reference file to ensure the test
        # passes for the current git checkout location and pytest's temp dir.
        # Using Jinja for this because other python str formatting methods require
        # too much manipulation of the reference files.
        environment = Environment(
            loader=FileSystemLoader(str(item.parent)),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = environment.get_template(item.name)
        # Note: jinja2' seems to be inconsistent with its trailing newlines depending
        # on the template and its if statements, so force a single trailing newline
        ref_text = template.render(**replacements).rstrip() + "\n"

        try:
            assert generated.exists()
            assert (
                ref_text == generated.open().read()
            ), "Reference does not match generated"
        except AssertionError:
            print("")
            print(f"Reference: {item}")
            print(f"Generated: {generated}")
            raise

    reference_paths = []

    def walk_dir(current):
        found_script = False
        for item in current.iterdir():
            if item.name == "spec.json":
                continue
            elif item.is_file():
                check_file(item)
                # At least one file was found beyond spec.json so test is valid
                found_script = True
            else:
                walk_dir(item)
            # Include the directory name so we match the rglob results below
            reference_paths.append(item)
        assert (
            found_script
        ), f"Reference dir needs to contain at least one script. {reference}"

    walk_dir(reference)

    # Check that we created the same number of files as the template
    processed_paths = sorted(Path(tmpdir).rglob("*"))
    reference_paths = sorted(reference_paths)

    total_processed = len(processed_paths)
    total_reference = len(reference_paths)
    assert (
        total_reference == total_processed
    ), "Reference file count doesn't match generated"

    # File paths should only be converted to the scripting language(cygpath)
    # only when generating a script, not before that. Check that the path was
    # not converted while building frozen data.
    # This check exists to ensure that we don't accidentally break this
    # requirement at some point in the future.
    freeze = cfg.freeze()
    for plat in ("linux", "windows"):
        alias = freeze["aliases"][plat]["inherited"]
        env = alias["environment"]["PATH"]
        distro = config_root / "distros" / "aliased" / "2.0"

        assert Path(alias["cmd"][1]) == distro / "list_vars.py"

        assert env[0] == "{PATH!e}"
        assert Path(env[1]) == distro / "PATH" / "env" / "with  spaces"


@pytest.mark.skip(reason="Find a way to test complex alias evaluation in pytest")
def test_complex_alias_bat(tmpdir, config_root):
    """This test is a placeholder for a future that can actually call hab's `hab.cli`
    and its aliases to check that they function correctly in Batch.

    This example text shows that using "hab env" can set an environment variable,
    and a complex alias call can modify that env var for the called command, but
    when the alias exits, the original env var is returned.
    """
    hab_paths = config_root / "site_main.json"
    test = (
        "REM Use the correct hab site configuration",
        f'set "HAB_PATHS={hab_paths}"',
        "REM activate the current hab environment",
        "hab env app/aliased/mod",
        'REM At this point the env var is set to "Global A"',
        'echo "In Hab env: " %ALIASED_GLOBAL_A%',
        "REM Call the alias, and check its output against the next section.",
        "REM This is showing that the alias process has modified env vars.",
        "global test",
        "REM After the alias was run, the global env var is restored",
        'echo "after calling function: " %ALIASED_GLOBAL_A%',
    )

    print("## Run these commands one at a time in a new bash terminal:")
    print("")
    print("\n".join(test))
    print("")
    print("## The output of calling global, should include:")
    print("- The `sys.argv:` line should include the test string passed to global")
    print(
        "- ALIASED_GLOBAL_A: ['Local Mod A', 'Local A Prepend', 'Global A', "
        "'Local A Append']"
    )
    raise AssertionError("Run this test manually line by line in a Command Prompt")


@pytest.mark.skip(reason="Find a way to test complex alias evaluation in pytest")
def test_complex_alias_ps1(tmpdir, config_root):
    """This test is a placeholder for a future that can actually call hab's `hab.cli`
    and its aliases to check that they function correctly in PowerShell.

    This example text shows that using "hab env" can set an environment variable,
    and a complex alias call can modify that env var for the called command, but
    when the alias exits, the original env var is returned.
    """
    hab_paths = config_root / "site_main.json"
    test = (
        "# Use the correct hab site configuration",
        f'$env:HAB_PATHS="{hab_paths}"',
        "# activate the current hab environment",
        "hab env app/aliased/mod",
        '# At this point the env var is set to "Global A"',
        'echo "In Hab env: " $env:ALIASED_GLOBAL_A',
        "# Call the alias, and check its output against the next section.",
        "# This is showing that the alias process has modified env vars.",
        "global test",
        "# After the alias was run, the global env var is restored",
        'echo "after calling function: " $env:ALIASED_GLOBAL_A',
    )

    print("## Run these commands one at a time in a new bash terminal:")
    print("")
    print("\n".join(test))
    print("")
    print("## The output of calling global, should include:")
    print("- The `sys.argv:` line should include the test string passed to global")
    print(
        "- ALIASED_GLOBAL_A: ['Local Mod A', 'Local A Prepend', 'Global A', "
        "'Local A Append']"
    )
    raise AssertionError("Run this test manually line by line in a PowerShell")


@pytest.mark.skip(reason="Find a way to test complex alias evaluation in pytest")
def test_complex_alias_sh(tmpdir, config_root):
    """This test is a placeholder for a future that can actually call hab's `hab.cli`
    and its aliases to check that they function correctly in Bash.

    This example text shows that using "hab env" can set an environment variable,
    and a complex alias call can modify that env var for the called command, but
    when the alias exits, the original env var is returned.
    """
    hab_paths = config_root / "site_main.json"
    test = (
        "# Use the correct hab site configuration",
        f'export HAB_PATHS="{hab_paths}"',
        "# activate the current hab environment",
        "hab env app/aliased/mod",
        '# At this point the env var is set to "Global A"',
        'echo "In Hab env: " $ALIASED_GLOBAL_A',
        "# Call the alias, and check its output against the next section.",
        "# This is showing that the alias process has modified env vars.",
        "global test",
        "# After the alias was run, the global env var is restored",
        'echo "after calling function: " $ALIASED_GLOBAL_A',
    )

    print("## Run these commands one at a time in a new bash terminal:")
    print("")
    print("\n".join(test))
    print("")
    print("## The output of calling global, should include:")
    print("- The `sys.argv:` line should include the test string passed to global")
    print(
        "- ALIASED_GLOBAL_A: ['Local Mod A', 'Local A Prepend', 'Global A', "
        "'Local A Append']"
    )
    raise AssertionError("Run this test manually line by line in a bash shell")


@pytest.mark.parametrize("ext", (".bat", ".ps1", ".sh"))
def test_invalid_alias(uncached_resolver, tmpdir, ext):
    """Check that useful errors are raised if an invalid alias is passed or if
    the alias doesn't have "cmd" defined.
    """
    kwargs = dict(ext=ext, exit=True, args=None, create_launch=True)

    # Check that calling a bad alias name raises a useful error message
    cfg = uncached_resolver.resolve("not_set/child")
    with pytest.raises(errors.HabError, match=r'"bad-alias" is not a valid alias name'):
        cfg.write_script(str(tmpdir), launch="bad-alias", **kwargs)

    # Remove the "cmd" value to test an invalid configuration
    alias = cfg.frozen_data["aliases"][utils.Platform.name()]["global"]
    del alias["cmd"]

    with pytest.raises(
        errors.HabError, match='Alias "global" does not have "cmd" defined'
    ):
        cfg.write_script(str(tmpdir), launch="global", **kwargs)



import functools
import os
import re
import site
import subprocess
import sys

import pytest

from hab import Resolver, Site, utils
from hab.errors import HabError


class Topen(subprocess.Popen):
    """A custom subclass of Popen."""


def missing_annotations_hack(function):
    """Decorator that works around a missing annotations module in python 3.6.

    Allows for calling subprocess calls in python 3.6 that would normally fail
    with a `SyntaxError: future feature annotations is not defined` exception.

    Works by temporarily removing the `_virtualenv.pth` file in the venv's
    site-packages.

    TODO: Figure out a better method until we can drop CentOS requirement.
    """

    @functools.wraps(function)
    def new_function(*args, **kwargs):
        if sys.version_info.minor != 6:
            return function(*args, **kwargs)

        site_packages = site.getsitepackages()
        for path in site_packages:
            pth = os.path.join(path, "_virtualenv.pth")
            if os.path.exists(pth):
                os.rename(pth, f"{pth}.bak")
        try:
            ret = function(*args, **kwargs)
        finally:
            for path in site_packages:
                pth = os.path.join(path, "_virtualenv.pth.bak")
                if os.path.exists(pth):
                    os.rename(pth, pth[:-3])
        return ret

    return new_function


def test_launch(resolver):
    """Check the Config.launch method."""
    cfg = resolver.resolve("app/aliased/mod")
    proc = cfg.launch("global", args=None, blocking=True)

    # Enabling mod_std sends all exception text to stdout
    assert proc.output_stderr is None
    check = [
        " Env Vars: ".center(80, "-"),
        "ALIASED_GLOBAL_A: ['Local Mod A', 'Local A Prepend', 'Global A', 'Local A Append']",
        "ALIASED_GLOBAL_B: ['Global B']",
        "ALIASED_GLOBAL_C: ['Local C Set']",
        "ALIASED_GLOBAL_D: <UNSET>",
        "ALIASED_GLOBAL_E: <UNSET>",
        "ALIASED_LOCAL: <UNSET>",
        "CONFIG_DEFINED: ['config_variable']",
        "",
        " PATH env var ".center(80, "-"),
    ]

    assert "\n".join(check) in proc.output_stdout


@missing_annotations_hack
def test_launch_str(resolver):
    cfg = resolver.resolve("app/aliased/mod")

    # Check that args are passed to the subprocess
    # Check that if "cmd" is a str, it is converted to a list and extended with args
    args = ["-c", 'print("success")']
    proc = cfg.launch("as_str", args=args, blocking=True)
    assert proc.output_stdout.strip() == "success"

    # Check that passing env is respected
    env = dict(os.environ)
    var_name = "TEST_ADDED_VARIABLE"
    var_value = "Test variable"
    assert var_name not in env
    env[var_name] = var_value

    # Check that Popen kwargs can be passed through launch, including env.
    args = [
        "-c",
        'import os;assert os.environ["{}"] != "{}"'.format(var_name, var_value),
    ]
    proc = cfg.launch("as_str", args=args, blocking=True, env=env)


@pytest.mark.skipif(sys.platform != "win32", reason="only applies on windows")
def test_pythonw(monkeypatch, resolver):
    """Check that sys.stdin is set if using pythonw."""
    cfg = resolver.resolve("app/aliased/mod")

    # If not using pythonw.exe proc.stdin is not routed to PIPE
    proc = cfg.launch("global", args=None, blocking=True)
    assert proc.stdin is None

    # If using pythonw.exe, proc.stdin is routed to PIPE
    pyw = re.sub("python.exe", "pythonw.exe", sys.executable, flags=re.I)
    monkeypatch.setattr(sys, "executable", pyw)
    proc = cfg.launch("global", args=None, blocking=True)
    assert proc.stdin is not None


def test_invalid_alias(resolver):
    cfg = resolver.resolve("app/aliased/mod")
    with pytest.raises(HabError, match='"not-a-alias" is not a valid alias name'):
        cfg.launch("not-a-alias")

    # Remove the "cmd" value to test an invalid configuration
    alias = cfg.frozen_data["aliases"][utils.Platform.name()]["global"]
    del alias["cmd"]

    with pytest.raises(HabError, match='Alias "global" does not have "cmd" defined'):
        cfg.launch("global")


def test_cls_no_entry_point(resolver):
    """Check that if no entry point is defined, `hab.launcher.Launcher` is
    used to launch the alias.
    """
    entry_points = resolver.site.entry_points_for_group("hab.launch_cls")
    assert len(entry_points) == 0

    cfg = resolver.resolve("app/aliased/mod")
    proc = cfg.launch("global", blocking=True)

    from hab.launcher import Launcher

    assert type(proc) is Launcher

    # Check that specifying cls, overrides default
    proc = cfg.launch("global", blocking=True, cls=Topen)
    assert type(proc) is Topen


def test_cls_entry_point(config_root):
    """Check that if an entry point is defined, it is used unless overridden."""
    site = Site(
        [config_root / "site/site_entry_point_a.json", config_root / "site_main.json"]
    )
    entry_points = site.entry_points_for_group("hab.launch_cls")
    assert len(entry_points) == 1
    # Test that the `test-gui` `hab.cli` entry point is handled correctly
    ep = entry_points[0]
    assert ep.name == "subprocess"
    assert ep.group == "hab.launch_cls"
    assert ep.value == "subprocess:Popen"

    resolver = Resolver(site=site)
    cfg = resolver.resolve("app/aliased/mod")

    # Check that entry_point site config is respected
    proc = cfg.launch("global", blocking=True)
    assert type(proc) is subprocess.Popen

    # Check that specifying cls, overrides site config
    proc = cfg.launch("global", blocking=True, cls=Topen)
    assert type(proc) is Topen


def test_alias_entry_point(config_root):
    """Check that if an entry point is defined on a complex alias, it is used."""
    site = Site(
        [config_root / "site/site_entry_point_a.json", config_root / "site_main.json"]
    )

    resolver = Resolver(site=site)
    cfg = resolver.resolve("app/aliased/mod")

    # NOTE: We need to compare the name of the classes because they are separate
    # imports that don't compare equal using `is`.

    # Check that entry_point site config is respected
    proc = cfg.launch("global", blocking=True)
    assert type(proc).__name__ == "Popen"

    # Check that if the complex alias specifies hab.launch_cls, it is used instead
    # of the site defined or default class.
    alias = cfg.frozen_data["aliases"][utils.Platform.name()]["global"]
    alias["hab.launch_cls"] = {"subprocess": "tests.test_launch:Topen"}
    proc = cfg.launch("global", blocking=True)
    assert type(proc).__name__ == "Topen"


@pytest.mark.parametrize("exit_code", [5, 4, 0])
class TestCliExitCodes:
    """Test that calling `hab launch` runs python and passes a complex command string
    correctly. Also checks that exit-codes are properly returned to the caller.
    """

    # Note: Using as_str to call python. I ran into lockups when trying to call
    # a "python" alias defined in the hab tests.
    sub_cmd = ["launch", "app/aliased", "as_str", "-c"]
    # This complex string is passed to python, and ensures that complex strings
    # get encoded correctly for the generated shell scripts. It also proves that
    # we are getting text output from python. The sys.exit call is not required,
    # but is used to ensure that exit-codes are returned to the calling process.
    py_cmd = "print('Running...');import sys;print(sys);sys.exit({code})"
    output_text = "Running...\n<module 'sys' (built-in)>\n"
    run_kwargs = dict(stdout=subprocess.PIPE, stderr=subprocess.STDOUT, timeout=5)
    if sys.version_info.minor >= 7:
        run_kwargs["text"] = True
    else:
        run_kwargs["universal_newlines"] = True

    @pytest.mark.skipif(sys.platform != "win32", reason="only applies on windows")
    @missing_annotations_hack
    def test_bat(self, config_root, exit_code):
        hab_bin = (config_root / ".." / "bin" / "hab.bat").resolve()
        # fmt: off
        cmd = [
            str(hab_bin), "--site", str(config_root / "site_main.json"),
            *self.sub_cmd,
            self.py_cmd.format(code=exit_code),
        ]
        # fmt: on

        # Run the hab command in a subprocess
        proc = subprocess.run(cmd, **self.run_kwargs)

        # Check that the print statement was actually run
        assert proc.stdout == self.output_text
        assert proc.returncode == exit_code

    @pytest.mark.skipif(sys.platform != "win32", reason="only applies on windows")
    @pytest.mark.skipif(
        os.getenv("GITHUB_ACTIONS") == "true",
        reason="PowerShell tests timeout when run in a github action",
    )
    @missing_annotations_hack
    def test_ps1(self, config_root, exit_code):
        # -File is needed to get the exit-code from powershell, and requires a full path
        script = (config_root / ".." / "bin" / "hab.ps1").resolve()
        # fmt: off
        cmd = [
            "powershell.exe", "-ExecutionPolicy", "Unrestricted",
            "-File", str(script),
            "--site", str(config_root / "site_main.json"),
            *self.sub_cmd,
            f"\"{self.py_cmd.format(code=exit_code)}\"",
        ]
        # fmt: on

        # Run the hab command in a subprocess
        proc = subprocess.run(cmd, **self.run_kwargs)

        # Check that the print statement was actually run
        assert proc.stdout == self.output_text
        assert proc.returncode == exit_code

    @pytest.mark.skip(reason="Find a way to make this test run successfully")
    def test_bash_win(self, config_root, exit_code):
        # fmt: off
        cmd = [
            "hab",
            "--site", f'{(config_root / "site_main.json").as_posix()}',
            *self.sub_cmd,
            self.py_cmd.format(code=exit_code),
        ]
        cmd = [
            # Note: This requires compatible with git bash, or the `--site` path
            # gets mangled.
            os.path.expandvars(r"%PROGRAMFILES%\Git\usr\bin\bash.exe"),
            "--login", "-c", subprocess.list2cmdline(cmd)
        ]
        # fmt: on

        # Run the hab command in a subprocess
        print(
            "# If you run this command in a command prompt the expected results "
            "are printed, but calling it here doesn't return any results."
        )
        print('# Make sure to check the exit-code with "echo %ERRORLEVEL%"')
        print(subprocess.list2cmdline(cmd))
        proc = subprocess.run(cmd, **self.run_kwargs)
        print(proc.stdout)
        print(proc.stderr)

        # Check that the print statement was actually run
        assert proc.stdout == self.output_text
        assert proc.returncode == exit_code

    @pytest.mark.skip(reason="Find a way to make this test run successfully")
    def test_bash_linux(self, config_root, exit_code):
        hab_bin = (config_root / ".." / "bin" / "hab").resolve()

        # fmt: off
        cmd = [
            hab_bin,
            "--site", f'{(config_root / "site_main.json")}',
            *self.sub_cmd,
            self.py_cmd.format(code=exit_code),
        ]
        # fmt: on

        # Run the hab command in a subprocess
        print(
            "# Run this command in a bash shell and verify that the expected results"
            " are printed. TODO: figure out how to run this with pytest."
        )
        print('# Make sure to check the exit-code with "echo $?"')
        print(subprocess.list2cmdline(cmd))
        proc = subprocess.run(cmd, shell=True, **self.run_kwargs)
        print(proc.stdout)
        print(proc.stderr)

        # Check that the print statement was actually run
        assert proc.stdout == self.output_text
        assert proc.returncode == exit_code



import os
import sys


def print_var(var):
    # Get the value from the environment
    value = os.getenv(var, "<UNSET>")
    if value != "<UNSET>":
        # Split it on path sep to ensure it is correctly generated
        value = value.split(os.path.pathsep)

    print("{}: {}".format(var, value))


print("sys.argv: {}".format(sys.argv))

# Print expected environment variable values
print(" Env Vars: ".center(80, "-"))
print_var("ALIASED_GLOBAL_A")
print_var("ALIASED_GLOBAL_B")
print_var("ALIASED_GLOBAL_C")
print_var("ALIASED_GLOBAL_D")
print_var("ALIASED_GLOBAL_E")
print_var("ALIASED_LOCAL")
print_var("CONFIG_DEFINED")
print("")

print(" PATH env var ".center(80, "-"))
for p in os.environ["PATH"].split(os.path.pathsep):
    print(p)

print("".center(80, "-"))



import copy
import json
import re
import sys
from pathlib import Path

import anytree
import pytest
import setuptools_scm
from packaging.version import Version

from hab import NotSet, utils
from hab.errors import (
    DuplicateJsonError,
    InvalidVersionError,
    ReservedVariableNameError,
    _IgnoredVersionError,
)
from hab.parsers import Config, DistroVersion, FlatConfig


def test_distro_parse(config_root, resolver):
    """Check that a distro json can be parsed correctly"""
    forest = {}
    app = DistroVersion(forest, resolver)
    path = config_root / "distros" / "all_settings" / "0.1.0.dev1" / ".hab.json"
    app.load(path)
    check = json.load(path.open())

    assert "{name}=={version}".format(**check) == app.name
    assert Version(check["version"]) == app.version
    assert check["environment"] == app.environment_config

    assert check["aliases"] == app.aliases
    assert ["all_settings"] == app.context

    # Verify that if the json file doesn't have "version" defined it uses the
    # parent directory as its version.
    app = DistroVersion(forest, resolver)
    path = config_root / "distros" / "maya2020" / "2020.0" / ".hab.json"
    app.load(path)
    check = json.load(path.open())

    # tests\distros\maya\2020.0\.hab.json does not have "version"
    # defined. This allows us to test that DistroVersion will pull the
    # version number from the parent directory not the json file.
    assert "version" not in check
    assert app.version == Version("2020.0")


def test_distro_version_resolve(config_root, resolver, helpers, monkeypatch, tmpdir):
    """Check the various methods for DistroVersion.version to be populated."""

    # Test that `.hab_version.txt` is respected if it exists.
    forest = {}
    app = DistroVersion(forest, resolver)
    path = config_root / "distros_version" / "txt_file" / ".hab.json"
    app.load(path)
    assert app.version == Version("1.7")

    # Test that an error is raised if the version could not be determined
    path = config_root / "distros_version" / "not_scm" / ".hab.json"
    with pytest.raises(InvalidVersionError, match=r"Hab was unable to determine"):
        app.load(path)

    # Test that an nice error is raised if setuptools_scm is not installed
    with monkeypatch.context() as m:
        # Simulate that setuptools-scm is not installed
        m.setitem(sys.modules, "setuptools_scm", None)
        with pytest.raises(
            InvalidVersionError,
            match=r"\[ModuleNotFoundError\] import of setuptools_scm halted",
        ):
            app.load(path)

    # Test that setuptools_scm is able to resolve the version.
    # This env var forces setuptools_scm to this version so we don't have to
    # create a git repo to test that get_version is called correctly.
    with monkeypatch.context() as m:
        m.setenv("SETUPTOOLS_SCM_PRETEND_VERSION", "1.9")
        app.load(path)
        assert app.version == Version("1.9")

    # Check that we add debug info to arbitrary setuptools_scm Exceptions.
    def get_ver(*args, **kwargs):
        raise OSError("Simulate a arbitrary error in setuptools_scm.get_version")

    with monkeypatch.context() as m:
        m.setattr(setuptools_scm, "get_version", get_ver)
        with pytest.raises(InvalidVersionError, match=r"\[OSError\] Simulate"):
            app.load(path)

    # Test that if the dirname matches `resolver.ignored`, the folder is
    # skipped by raising an _IgnoredVersionError exception.
    path = config_root / "distros_version" / "release" / ".hab.json"
    with pytest.raises(
        _IgnoredVersionError, match=r"its dirname is in the ignored list"
    ):
        app.load(path)

    # Check that `resolver.ignored` is also respected for arbitrary
    # setuptools_scm Exceptions.
    with monkeypatch.context() as m:
        m.setattr(setuptools_scm, "get_version", get_ver)
        with pytest.raises(
            _IgnoredVersionError, match=r"its dirname is in the ignored list"
        ):
            app.load(path)


def test_distro_version(resolver):
    """Verify that we find the expected version for a given requirement."""
    maya = resolver.distros["maya2020"]

    assert maya.latest_version("maya2020").name == "maya2020==2020.1"
    assert maya.latest_version("maya2020<2020.1").name == "maya2020==2020.0"


def test_config_parse(config_root, resolver, helpers):
    """Check that a config json can be parsed correctly"""
    forest = {}
    config = Config(forest, resolver)
    path = config_root / "configs" / "default" / "default.json"
    config.load(path)

    check = [
        "maya2020",
        "the_dcc_plugin_a",
        "the_dcc_plugin_b",
        "the_dcc_plugin_c",
        "houdini18.5",
        "the_dcc_plugin_d",
    ]

    # We can't do a simple comparison of Requirement keys so check that these resolved
    helpers.assert_requirements_equal(config.distros, check)

    # Check that the forest was populated correctly
    assert set(forest.keys()) == set(["default"])
    assert forest["default"] == config


def test_config_parenting(config_root, resolver):
    """Check that a correct tree structure is generated especially when loaded
    in a incorrect order.
    """

    def repr_list(node):
        return [repr(x) for x in anytree.iterators.PreOrderIter(node)]

    forest = {}
    root_paths = set((config_root,))
    # Ensure the forest has multiple trees when processing
    shared_path = config_root / "configs" / "default" / "default.json"
    Config(forest, resolver, filename=shared_path, root_paths=root_paths)

    # Load the tree structure from child to parent to test that the placeholder system
    # works as expected
    Config(
        forest,
        resolver,
        filename=config_root
        / "configs"
        / "project_a"
        / "project_a_Sc001_animation.json",
        root_paths=root_paths,
    )
    check = [
        "hab.parsers.placeholder.Placeholder('project_a')",
        "hab.parsers.placeholder.Placeholder('project_a/Sc001')",
        "hab.parsers.config.Config('project_a/Sc001/Animation')",
    ]
    assert check == repr_list(forest["project_a"])

    # Check that a middle plcaeholder was replaced
    mid_level_path = config_root / "configs" / "project_a" / "project_a_Sc001.json"
    Config(forest, resolver, filename=mid_level_path, root_paths=root_paths)
    check[1] = "hab.parsers.config.Config('project_a/Sc001')"
    assert check == repr_list(forest["project_a"])

    # Check that a middle Config object is used not replaced
    Config(
        forest,
        resolver,
        filename=config_root / "configs" / "project_a" / "project_a_Sc001_rigging.json",
        root_paths=root_paths,
    )
    check.append("hab.parsers.config.Config('project_a/Sc001/Rigging')")
    assert check == repr_list(forest["project_a"])

    # Check that a root item is replaced
    top_level_path = config_root / "configs" / "project_a" / "project_a.json"
    Config(forest, resolver, filename=top_level_path, root_paths=root_paths)
    check[0] = "hab.parsers.config.Config('project_a')"
    assert check == repr_list(forest["project_a"])

    # Verify that the correct exceptions are raised if root duplicates are loaded
    with pytest.raises(DuplicateJsonError):
        Config(forest, resolver, filename=top_level_path, root_paths=root_paths)
    assert check == repr_list(forest["project_a"])

    # and at the leaf level
    with pytest.raises(DuplicateJsonError):
        Config(forest, resolver, filename=mid_level_path, root_paths=root_paths)
    assert check == repr_list(forest["project_a"])

    # Check that the forest didn't loose the default tree
    assert ["hab.parsers.config.Config('default')"] == repr_list(forest["default"])


def test_metaclass():
    assert set(DistroVersion._properties.keys()) == set(
        [
            "alias_mods",
            "aliases",
            "distros",
            "environment",
            "environment_config",
            "filename",
            "min_verbosity",
            "name",
            "variables",
            "version",
        ]
    )
    assert set(Config._properties.keys()) == set(
        [
            "alias_mods",
            "aliases",
            "distros",
            "environment",
            "environment_config",
            "filename",
            "min_verbosity",
            "inherits",
            "name",
            "uri",
            "variables",
        ]
    )


class TestDump:
    def test_dump(self, resolver):
        line = "-LINE-"

        def standardize(txt):
            """Dump adds an arbitrary length border of -'s that makes str comparing
            hard. Replace it with a constant value."""
            return re.sub(r"^-+$", line, txt, flags=re.M)

        # Build the test data so we can generate the output to check
        # Note: using `repr([u"` so this test passes in python 2 and 3
        pre = ["name:  child", "uri:  not_set/child"]
        post = ["inherits:  True", "min_verbosity:  NotSet"]
        env = [
            "environment:  FMT_FOR_OS:  a{;}b;c:{PATH!e}{;}d",
            "              TEST:  case",
            "              UNSET_VARIABLE:  None",
        ]

        env_config = [
            "environment_config:  set:  FMT_FOR_OS:  a{;}b;c:{PATH!e}{;}d",
            "                           TEST:  case",
            "                     unset:  UNSET_VARIABLE",
        ]
        cfg = resolver.closest_config("not_set/child")
        header = f"Dump of {type(cfg).__name__}('{cfg.fullpath}')"

        # Check that both environments can be hidden
        result = cfg.dump(
            environment=False, environment_config=False, verbosity=2, color=False
        )
        check = [f"{header}\n{line}"]
        check.extend(pre)
        check.extend(post)
        check.append(line)
        check = "\n".join(check)
        assert standardize(result) == check

        # Check that both environments can be shown
        result = cfg.dump(
            environment=True, environment_config=True, verbosity=2, color=False
        )
        check = [f"{header}\n{line}"]
        check.extend(pre)
        check.extend(env)
        check.extend(env_config)
        check.extend(post)
        check.append(line)
        check = "\n".join(check)
        assert standardize(result) == check

        # Check that only environment can be shown
        result = cfg.dump(
            environment=True, environment_config=False, verbosity=2, color=False
        )
        check = [f"{header}\n{line}"]
        check.extend(pre)
        check.extend(env)
        check.extend(post)
        check.append(line)
        check = "\n".join(check)
        # Check that only environment_config can be shown
        result = cfg.dump(
            environment=False, environment_config=True, verbosity=2, color=False
        )
        check = [f"{header}\n{line}"]
        check.extend(pre)
        check.extend(env_config)
        check.extend(post)
        check.append(line)
        check = "\n".join(check)
        assert standardize(result) == check

    def test_flat(self, resolver):
        """Test additional dump settings for FlatConfig objects"""
        cfg = resolver.resolve("not_set/child")
        # Check that dump formats versions nicely
        check = re.compile(
            r"versions:  (?P<ver>aliased==2\.0, maya2020==2020\.1)"
            r"(?P<file>:  [\w:\\.\/-]+\.json)?"
        )
        # Versions are not shown with verbosity >= 1
        result = cfg.dump(color=False)
        assert not check.search(result)

        # Verbosity 2 shows just the version requirement
        result = cfg.dump(verbosity=2, color=False)
        match = check.search(result)
        assert match.group("file") is None
        assert match.group("ver") == "aliased==2.0, maya2020==2020.1"

        # Verbosity 3 also shows the json file name
        check = re.compile(
            r"versions:  (?P<vera>aliased==2\.0)(?P<filea>:  [\w:\\.\/-]+\.json)?\n"
            r"           (?P<verm>maya2020==2020\.1)(?P<filem>:  [\w:\\.\/-]+\.json)?"
        )
        result = cfg.dump(verbosity=3, color=False)
        match = check.search(result)
        assert match.group("filea") is not None
        assert match.group("filem") is not None

    @pytest.mark.parametrize("uri", ("not_set/no_distros", "not_set/empty_lists"))
    def test_no_values(self, resolver, uri):
        """If noting sets aliases or distros, don't include them in the dump text."""
        cfg = resolver.resolve(uri)

        for verbosity in range(3):
            result = cfg.dump(verbosity=verbosity, color=False)
            assert "aliases" not in result
            assert "distros:" not in result


def test_environment(resolver):
    # Check that the correct errors are raised
    cfg = resolver.closest_config("not_set/env_path_set")
    with pytest.raises(
        ValueError, match=r'You can not use PATH for the set operation: "path_variable"'
    ):
        cfg.environment

    cfg = resolver.closest_config("not_set/env_path_unset")
    with pytest.raises(ValueError, match=r"You can not unset PATH"):
        cfg.environment

    # Attempting to use a reserved env var raises an exception
    # Note: KeyError always adds quotes around the message passed so we need to
    # add them when checking the exception text
    # https://stackoverflow.com/a/24999035
    cfg = resolver.closest_config("not_set/env_path_hab_uri")
    with pytest.raises(
        KeyError, match=r"'\"HAB_URI\" is a reserved environment variable'"
    ):
        cfg.environment

    # Check environment variable resolving
    cfg = resolver.closest_config("not_set/env1")

    assert cfg.environment["APPEND_VARIABLE"] == ["append_value"]
    assert cfg.environment["MAYA_MODULE_PATH"] == ["MMP_Set"]
    assert cfg.environment["PREPEND_VARIABLE"] == ["prepend_value"]

    check = [
        "{relative_root}/prepend",
        "{relative_root}/set",
        "{relative_root}/append",
    ]
    relative_root = utils.path_forward_slash(cfg.dirname)
    check = [c.format(relative_root=relative_root) for c in check]

    assert cfg.environment["RELATIVE_VARIABLE"] == check
    assert cfg.environment["SET_RELATIVE"] == [
        "{relative_root}".format(relative_root=relative_root)
    ]
    assert cfg.environment["SET_VARIABLE"] == ["set_value"]
    assert cfg.environment["UNSET_VARIABLE"] is None
    assert cfg.environment["UNSET_VARIABLE_1"] is None

    # Check `cfg.environment is NotSet` resolves correctly
    cfg = resolver.closest_config("not_set")
    assert cfg.environment == {}

    # Ensure our tests cover the early out if the config is missing append/prepend
    cfg = resolver.closest_config("not_set/child")
    assert cfg.environment == {
        "TEST": ["case"],
        "FMT_FOR_OS": ["a{;}b;c:{PATH!e}{;}d"],
        "UNSET_VARIABLE": None,
    }


def test_flat_config(resolver):
    ret = resolver.resolve("not_set/child")
    check = {
        "TEST": ["case"],
        "FMT_FOR_OS": ["a{;}b;c:{PATH!e}{;}d"],
        "UNSET_VARIABLE": None,
        "HAB_URI": ["not_set/child"],
        "ALIASED_GLOBAL_A": ["Global A"],
        "ALIASED_GLOBAL_B": ["Global B"],
        "ALIASED_GLOBAL_C": ["Global C"],
        "ALIASED_GLOBAL_D": ["Global D"],
        "ALIASED_GLOBAL_E": None,
        "ALIASED_GLOBAL_F": ["Global F"],
    }

    assert ret.environment == check
    assert ret.environment == check
    assert ret.environment == check

    # Check for edge case where self._environment was reset if the config didn't define
    # environment, but the attached distros did.
    ret = resolver.resolve("not_set/no_env")
    assert sorted(ret.environment.keys()) == [
        "DCC_CONFIG_PATH",
        "DCC_MODULE_PATH",
        "HAB_URI",
    ]
    assert sorted(ret.environment.keys()) == [
        "DCC_CONFIG_PATH",
        "DCC_MODULE_PATH",
        "HAB_URI",
    ]


def test_flat_config_env_resolve(resolver, config_root, helpers):
    """Checks that environment variables are properly merged when resolving distros
    including inherited values. Checks that append/prepend are processed correctly
    and result in a consistent ordering of the resulting environment variables.
    """
    ret = resolver.resolve("not_set/distros")

    # Ensure the the configuration files this test relies on are configured
    # correctly. This also serves as a explanation for why the final list being
    # checked is sorted in the way it is
    raw_json = json.load(open(ret.filename))
    # The config only links to the "the_dcc" distro and no others.
    assert raw_json["distros"] == {"the_dcc": []}

    # The_dcc depends on these three distros in this order
    distro_root = config_root / "distros"
    raw_json = json.load((distro_root / "the_dcc" / "1.2" / ".hab.json").open())
    assert "the_dcc_plugin_a" in raw_json["distros"][0]
    assert "the_dcc_plugin_b" in raw_json["distros"][1]
    assert "the_dcc_plugin_e" in raw_json["distros"][2]

    # the_dcc_plugin_a depends on these two distros in this order. The dependencies
    # are resolved down the tree so "e" will show up before "b".
    distro_root = config_root / "distros"
    raw_json = json.load(
        (distro_root / "the_dcc_plugin_a" / "1.1" / ".hab.json").open()
    )
    assert "the_dcc_plugin_e" in raw_json["distros"][0]
    assert "the_dcc_plugin_d" in raw_json["distros"][1]

    # Both environment variables are appended
    for plugin in ("the_dcc_plugin_a", "the_dcc_plugin_e"):
        raw_json = json.load((distro_root / plugin / "1.1" / ".hab.json").open())
        assert "DCC_CONFIG_PATH" in raw_json["environment"]["append"]
        assert "DCC_MODULE_PATH" in raw_json["environment"]["append"]

    # One env var is appended and the other is prepended
    for plugin in ("the_dcc_plugin_b", "the_dcc_plugin_d"):
        raw_json = json.load((distro_root / plugin / "1.1" / ".hab.json").open())
        assert "DCC_CONFIG_PATH" in raw_json["environment"]["prepend"]
        assert "DCC_MODULE_PATH" in raw_json["environment"]["append"]

    # Check that the environment was actually processed depth-frist, the above
    # serves as documentation for why these env vars are ordered the way they are

    # Ensure that appends are handled correctly
    helpers.check_path_list(
        ret.environment["DCC_MODULE_PATH"],
        [
            # plugin a was first referenced so it is the first append
            str(config_root / "distros" / "the_dcc_plugin_a" / "1.1"),
            # plugin e was first in the distros specified by plugin a, so it
            # gets added next even though the_dcc referenced plugin b first
            str(config_root / "distros" / "the_dcc_plugin_e" / "1.1"),
            # plugin d was also referenced by plugin a, so it goes before plugin b
            str(config_root / "distros" / "the_dcc_plugin_d" / "1.1"),
            # Finally plugin b is added when the dependency resolver finishes
            # processing plugin a and returns to the next distro for the_dcc.
            str(config_root / "distros" / "the_dcc_plugin_b" / "1.1"),
            # plugin e was already added by plugin a, so it doesn't get re-added
            # when the dependency resolver finds it in the_dcc's distros.
        ],
    )

    # Ensure that prepends are handled correctly
    helpers.check_path_list(
        ret.environment["DCC_CONFIG_PATH"],
        [
            # plugin b is processed last, so it's prepend ends up first
            str(config_root / "distros" / "the_dcc_plugin_b" / "1.1"),
            # plugin d is processed before b, so it's prepend is next
            str(config_root / "distros" / "the_dcc_plugin_d" / "1.1"),
            # The rest are appends, and follow the same order as above
            str(config_root / "distros" / "the_dcc_plugin_a" / "1.1"),
            str(config_root / "distros" / "the_dcc_plugin_e" / "1.1"),
        ],
    )


def test_placeholder_handling(resolver):
    # We haven't defined the top level "placeholder" uri or "placeholder/undefined",
    # but we have defined "placeholder/child". Check that the Placeholder objects
    # are processed correctly. Ie in this case it inherits from the default config.
    ret = resolver.resolve("place-holder/undefined")
    # Name and context are set for every Placeholder
    assert ret.name == "place-holder"
    assert ret.context is NotSet
    assert ret.fullpath == "place-holder"

    # Verify that default configuration settings were loaded because inherits
    # defaults to True.
    assert "the_dcc" not in ret.distros.keys()
    assert "maya2020" in ret.distros.keys()
    assert "the_dcc_plugin_a" in ret.distros.keys()
    # HAB_URI is always added to the environment variables
    assert ret.environment == {"HAB_URI": ["place-holder/undefined"]}

    # Check that if inherits is False, inheritance doesn't happen with Placeholders.
    ret = resolver.resolve("place-holder/child")
    # Name and context are set for every Placeholder
    assert ret.name == "child"
    assert ret.context == ["place-holder"]
    assert ret.fullpath == "place-holder/child"

    assert "the_dcc" in ret.distros
    assert "maya2020" not in ret.distros
    assert "the_dcc_plugin_a" not in ret.distros
    assert len(ret.environment) == 3
    assert "DCC_CONFIG_PATH" in ret.environment
    assert "DCC_MODULE_PATH" in ret.environment
    # HAB_URI is always added to the environment variables
    assert "HAB_URI" in ret.environment

    # Check that if inherits is True, inheritance happens with Placeholders.
    ret = resolver.resolve("place-holder/inherits")
    # Name and context are set for every Placeholder
    assert ret.name == "inherits"
    assert ret.context == ["place-holder"]
    assert ret.fullpath == "place-holder/inherits"

    # Inherits does not define distros, so it inherits distros from default.
    assert "the_dcc" not in ret.distros.keys()
    assert "maya2020" in ret.distros.keys()
    assert "the_dcc_plugin_a" in ret.distros.keys()
    # Inherits defines environment, so its environment settings are not inherited.
    assert len(ret.environment) == 2
    assert "DCC_MODULE_PATH" not in ret.environment
    # HAB_URI is always added to the environment variables
    assert "HAB_URI" in ret.environment
    assert "TEST" in ret.environment


def test_invalid_config(config_root, resolver):
    """Check that if an invalid json file is processed, its filename is included in
    the traceback"""
    path = config_root / "invalid.json"

    # Resolve if we are using pyjson5 or the native json class
    native_json = False
    try:
        from pyjson5 import Json5Exception as _JsonException
    except ImportError:
        native_json = True
        from builtins import ValueError as _JsonException

    with pytest.raises(_JsonException) as excinfo:
        Config({}, resolver, filename=path)

    if native_json:
        # If built-in json was used, check that filename was appended to the message
        assert f'Filename("{path}")' in str(excinfo.value)
    else:
        # If pyjson5 was used, check that the filename was added to the result dict
        assert excinfo.value.result["filename"] == str(path)


def test_misc_coverage(resolver):
    """Test that cover misc lines not covered by the rest of the tests"""
    assert str(NotSet) == "NotSet"
    assert copy.copy(NotSet) is NotSet

    # Check that dirname is modified when setting a blank filename
    cfg = Config({}, resolver)
    cfg.filename = ""
    # both of these values should be set to ``Path(os.devnull)``
    assert str(cfg.filename) in ("nul", "/dev/null")
    assert str(cfg.dirname) in ("nul", "/dev/null")

    # String values are cast to the correct type
    cfg.version = "1.0"
    assert cfg.version == Version("1.0")

    # Loading a directory raises a FileNotFoundError
    with pytest.raises(FileNotFoundError):
        cfg.load(".")

    # Loading a non-existent file path raises a FileNotFoundError
    with pytest.raises(FileNotFoundError):
        cfg.load("invalid_path.json")


@pytest.mark.parametrize(
    "dirname,uri",
    (
        ("configs_parent", "not_set"),
        ("configs_child", "not_set/child"),
    ),
)
def test_duplicated_configs(config_root, resolver, dirname, uri):
    """Check that a specific config can be duplicated and that the first path
    is the used data. If there is a duplicate in the same config path, an exception
    is raised.

    The duplicates/*_1 folders have a second definition, but are in their own unique
    config_paths, so only the first found config is used.

    The duplicates/*_2 folders have a third definition. The second and third definitions
    are in the same config_path so a DuplicateJsonError is raised.
    """
    original = resolver.config_paths
    config_paths = list(original)
    # Check that config_paths raise exceptions correctly
    config_paths.insert(0, config_root / "duplicates" / f"{dirname}_1")
    resolver.config_paths = config_paths

    # Check that the first config in config_paths was used
    cfg = resolver.resolve(uri)
    assert f"{dirname}_1" in str(cfg.filename)

    # Check that an exception is raised if there are duplicate definitions from
    # the same config_paths directory.
    config_paths = list(original)
    config_paths.insert(0, config_root / "duplicates" / f"{dirname}_2")
    resolver.config_paths = config_paths
    with pytest.raises(DuplicateJsonError):
        resolver.resolve(uri)


def test_duplicated_distros(config_root, resolver):
    """Check that a specific config can be duplicated and that the first path
    is the used data. If there is a duplicate in the same config path, an exception
    is raised.

    The duplicates/distros_1 folders have a second definition, but are in their own
    unique config_paths, so only the first found config is used.

    The duplicates/distros_2 folders have a third definition. The second and third
    definitions are in the same config_path so a DuplicateJsonError is raised.
    """
    original = resolver.distro_paths

    # Check that the first config in distro_paths was used
    distro_paths = list(original)
    distro_paths.insert(0, config_root / "duplicates" / "distros_1" / "*")
    resolver.distro_paths = distro_paths

    dcc = resolver.find_distro("the_dcc==1.2")
    assert dcc.name == "the_dcc==1.2"
    assert "distros_1" in str(dcc.filename)

    # Check that an exception is raised if there are duplicate definitions from
    # the same distro_paths directory.
    distro_paths = list(original)
    distro_paths.insert(0, config_root / "duplicates" / "distros_2" / "*")
    resolver.distro_paths = distro_paths

    with pytest.raises(DuplicateJsonError):
        resolver.find_distro("the_dcc==1.2")


def test_os_specific_linux(monkeypatch, resolver):
    """Check that if "os_specific" is set to true, only env vars for the current
    os are resolved."""
    # Simulate running on a linux platform.
    monkeypatch.setattr(utils, "Platform", utils.LinuxPlatform)
    cfg = resolver.resolve("not_set/os")

    assert cfg.environment["UNSET_VARIABLE_LIN"] is None
    assert cfg.environment["SET_VARIABLE_LIN"] == ["set_value_lin"]
    assert cfg.environment["APPEND_VARIABLE_LIN"] == ["append_value_lin"]
    assert cfg.environment["PREPEND_VARIABLE_LIN"] == ["prepend_value_lin"]


def test_os_specific_win(monkeypatch, resolver):
    """Check that if "os_specific" is set to true, only env vars for the current
    os are resolved."""
    # Simulate running on a windows platform
    monkeypatch.setattr(utils, "Platform", utils.WinPlatform)
    cfg = resolver.resolve("not_set/os")

    assert cfg.environment["UNSET_VARIABLE_WIN"] is None
    assert cfg.environment["SET_VARIABLE_WIN"] == ["set_value_win"]
    assert cfg.environment["APPEND_VARIABLE_WIN"] == ["append_value_win"]
    assert cfg.environment["PREPEND_VARIABLE_WIN"] == ["prepend_value_win"]


class TestAliasMods:
    def test_global(self, resolver):
        """Check the various operations that alias_mod creates."""
        cfg = resolver.resolve("app/aliased/mod")
        alias = cfg.aliases["global"]

        # Check global hab managed env vars
        assert cfg.environment["ALIASED_GLOBAL_A"] == ["Global A"]
        assert cfg.environment["ALIASED_GLOBAL_B"] == ["Global B"]
        assert cfg.environment["ALIASED_GLOBAL_C"] == ["Global C"]
        assert cfg.environment["ALIASED_GLOBAL_D"] == ["Global D"]
        assert cfg.environment["ALIASED_GLOBAL_E"] is None
        assert cfg.environment["ALIASED_GLOBAL_F"] == ["Global F"]
        assert cfg.environment["ALIASED_MOD_GLOBAL_A"] == ["Global Mod A"]
        # This global env var was defined by the config json file.
        # Ensure that it did not get lost at some point in the process.
        assert cfg.environment["CONFIG_DEFINED"] == ["config_variable"]
        # This variable is always added by hab automatically
        assert cfg.environment["HAB_URI"] == ["app/aliased/mod"]
        # Ensure no extra env vars were defined
        assert len(cfg.environment) == 9

        # Check cmd is expected
        assert alias["cmd"][0] == "python"
        assert alias["cmd"][1].endswith("list_vars.py")

        env = alias["environment"]
        # Check that order of prepend/append env vars is correct from all sources
        # Order of processing of env vars: Global hab, alias, alias_mod.
        assert env["ALIASED_GLOBAL_A"] == [
            # 4. Prepended alias_mod env var processed last
            "Local Mod A",
            # 2. Prepend alias defined env var
            "Local A Prepend",
            # 1. Hab defined global env vars are added first
            "Global A",
            # 3. Append alias defined env var
            "Local A Append",
        ]
        # Set by alias env var(replace value)
        assert env["ALIASED_GLOBAL_C"] == ["Local C Set"]
        # This variable is unset for this alias(removing global value)
        assert env["ALIASED_GLOBAL_D"] is None
        # Set globally, and is being prepended by alias_mod
        assert env["ALIASED_GLOBAL_F"] == ["Local Mod F", "Global F"]
        # Set globally by aliased_mod, modified by aliased_mod
        assert env["ALIASED_MOD_LOCAL_B"] == ["Local Mod B"]

    def test_as_list(self, resolver):
        """Check that non-dict defined aliases are modified correctly"""
        cfg = resolver.resolve("app/aliased/mod")
        alias = cfg.aliases["global"]

        # Check cmd is expected
        assert alias["cmd"][0] == "python"
        assert alias["cmd"][1].endswith("list_vars.py")

        # Env var can be set by alias_mod without any other hab management
        assert alias["environment"]["ALIASED_MOD_LOCAL_B"] == ["Local Mod B"]

    def test_as_dict(self, resolver, config_root):
        """Additional tests for alias_mod."""
        cfg = resolver.resolve("app/aliased/mod")
        alias = cfg.aliases["as_dict"]

        # Check cmd is expected
        assert alias["cmd"][0] == "python"
        assert alias["cmd"][1].endswith("list_vars.py")

        env = alias["environment"]

        # Check alias_mod is prepended to alias defined env var
        # Also check that the alias_mods {relative_root} path is resolved to the directory
        assert (
            Path(env["ALIASED_LOCAL"][0])
            == config_root / "distros/aliased_mod/1.0/modified"
        )
        # Check that the alias {relative_root} path is resolved to it's directory
        assert Path(env["ALIASED_LOCAL"][1]) == config_root / "distros/aliased/2.0/test"
        # Env var can be set by alias_mod without any other hab management
        assert env["ALIASED_MOD_LOCAL_A"] == ["Local Mod A"]

    def test_as_dict_config(self, resolver, config_root):
        """Additional test for alias_mod being set by a config not a distro."""
        cfg = resolver.resolve("app/aliased/config")
        alias = cfg.aliases["as_dict"]
        env = alias["environment"]

        # Check alias_mod is prepended to alias defined env var and that {relative_root}
        # is pointing to the config directory
        assert Path(env["ALIASED_LOCAL"][0]) == config_root / "configs/app/config"
        # Check that the alias {relative_root} path is resolved to it's directory
        assert Path(env["ALIASED_LOCAL"][1]) == config_root / "distros/aliased/2.0/test"
        # Check that the config's alias_mod env var was added
        assert env["ALIASED_MOD_LOCAL_A"] == ["Local Config A"]

    def test_as_dict_mod_config(self, resolver, config_root):
        """Additional test for alias_mod when set by both config and distro.
        In this case the config's mods are processed before the distro mod. This
        means that the distro's changes wrap the config changes.
        """
        cfg = resolver.resolve("app/aliased/mod/config")
        alias = cfg.aliases["as_dict"]
        env = alias["environment"]

        # Check that the first alias_mod is the distro and relative to the distro
        assert (
            Path(env["ALIASED_LOCAL"][0])
            == config_root / "distros/aliased_mod/1.0/modified"
        )
        # Next is the config alias_mod and it is relative to the config
        assert Path(env["ALIASED_LOCAL"][1]) == config_root / "configs/app/config_mod"
        # Check that the alias {relative_root} path is resolved to it's directory
        assert Path(env["ALIASED_LOCAL"][2]) == config_root / "distros/aliased/2.0/test"
        # Check that the distro then config's env vars were both added in order
        assert env["ALIASED_MOD_LOCAL_A"] == ["Local Mod A", "Local Config Mod A"]


def test_duplicates(resolver):
    """Ensure consistent handling of duplicate alias names."""

    # houdini18.5 is the first distro, it's duplicate generic alias is used
    cfg = resolver.resolve("app/houdini/a")
    assert "18.5" in cfg.aliases["houdini"]["cmd"]
    assert "18.5" in cfg.aliases["houdini18.5"]["cmd"]
    assert "19.5" in cfg.aliases["houdini19.5"]["cmd"]

    # houdini19.5 is the first distro, it's duplicate generic alias is used
    cfg = resolver.resolve("app/houdini/b")
    assert "19.5" in cfg.aliases["houdini"]["cmd"]
    assert "18.5" in cfg.aliases["houdini18.5"]["cmd"]
    assert "19.5" in cfg.aliases["houdini19.5"]["cmd"]


def test_get_min_verbosity(resolver):
    cfg = resolver.resolve("verbosity")
    # Default is returned if an invalid config is passed
    assert cfg.get_min_verbosity(None, target="test", default=99) == 99
    # Default is returned if config's min_verbosity is NotSet
    data = {"min_verbosity": NotSet}
    assert cfg.get_min_verbosity(data, target="test", default=99) == 99
    # The requested target value is returned
    data["min_verbosity"] = {"global": 1, "test": 2, "hab": 3}
    assert cfg.get_min_verbosity(data, target="test", default=99) == 2
    assert cfg.get_min_verbosity(data, target="hab", default=99) == 3
    # The global value is returned if target is un-defined
    assert cfg.get_min_verbosity(data, target="missing", default=99) == 1


def test_alias_min_verbosity_default(resolver):
    """Test dumping of configs using the default target `hab`."""
    cfg = resolver.resolve("verbosity/inherit")

    with utils.verbosity_filter(resolver, verbosity=None):
        result = cfg.aliases
        assert sorted(result.keys()) == ["vb0", "vb1", "vb2", "vb3", "vb_default"]

    with utils.verbosity_filter(resolver, verbosity=2):
        result = cfg.aliases
        assert sorted(result.keys()) == ["vb0", "vb1", "vb2", "vb_default"]

    with utils.verbosity_filter(resolver, verbosity=1):
        result = cfg.aliases
        assert sorted(result.keys()) == ["vb0", "vb1", "vb_default"]

    with utils.verbosity_filter(resolver, verbosity=0):
        result = cfg.aliases
        assert sorted(result.keys()) == ["vb0", "vb_default"]


def test_alias_min_verbosity_hab_gui(resolver):
    """Test dumping of aliases using the non-default target `hab-gui`.."""
    cfg = resolver.resolve("verbosity/inherit")

    with utils.verbosity_filter(resolver, verbosity=None, target="hab-gui"):
        result = cfg.aliases
        assert sorted(result.keys()) == ["vb0", "vb1", "vb2", "vb3", "vb_default"]

    with utils.verbosity_filter(resolver, verbosity=2, target="hab-gui"):
        result = cfg.aliases
        assert sorted(result.keys()) == ["vb1", "vb2", "vb3", "vb_default"]

    with utils.verbosity_filter(resolver, verbosity=1, target="hab-gui"):
        result = cfg.aliases
        assert sorted(result.keys()) == ["vb2", "vb3", "vb_default"]

    with utils.verbosity_filter(resolver, verbosity=0, target="hab-gui"):
        result = cfg.aliases
        assert sorted(result.keys()) == ["vb3", "vb_default"]


def test_update_environ(resolver):
    pathsep = utils.Platform.pathsep(utils.Platform.default_ext())
    cfg = resolver.resolve("app/aliased/mod")

    # Pre-existing env var values. These are not managed by hab persist
    inherited = {
        "IGNORED": "This variable not managed by hab and is ignored",
    }

    # The global env vars that the hab config manages.
    check_global = {
        "ALIASED_GLOBAL_A": "Global A",
        "ALIASED_GLOBAL_B": "Global B",
        "ALIASED_GLOBAL_C": "Global C",
        "ALIASED_GLOBAL_D": "Global D",
        "ALIASED_GLOBAL_F": "Global F",
        "ALIASED_MOD_GLOBAL_A": "Global Mod A",
        "CONFIG_DEFINED": "config_variable",
        "HAB_URI": "app/aliased/mod",
    }
    # Include the inherited values
    check_global.update(inherited)

    def new_dict():
        return dict(inherited)

    def check_freeze(env):
        # Ensure the HAB_FREEZE variable was actually set on env and remove
        # it to make checking the rest of the dict easier.
        freeze = env.pop("HAB_FREEZE")
        # Check that it starts with the version identifier.
        assert re.match(r"v\d*:.+", freeze)

    # Check that global env vars were added
    env = new_dict()
    cfg.update_environ(env)
    check_freeze(env)
    assert env == check_global

    # And that variables are removed that are set to unset
    env = {"ALIASED_GLOBAL_E": "To be removed"}
    env.update(inherited)
    cfg.update_environ(env)
    check_freeze(env)
    assert env == check_global

    # Check aliased env vars are passed excluding global hab env vars
    # Alias defines no env vars
    env = new_dict()
    cfg.update_environ(env, alias_name="as_str", include_global=False)
    assert env == inherited

    # Alias output checking
    check_alias = {
        "ALIASED_GLOBAL_A": pathsep.join(
            ["Local Mod A", "Local A Prepend", "Global A", "Local A Append"]
        ),
        "ALIASED_GLOBAL_C": "Local C Set",
        "ALIASED_GLOBAL_F": pathsep.join(["Local Mod F", "Global F"]),
        "ALIASED_MOD_LOCAL_B": "Local Mod B",
    }
    # Include the inherited values
    check_alias.update(inherited)

    # Alias "global" defines several env vars, these include hab global env vars.
    env = new_dict()
    cfg.update_environ(env, alias_name="global", include_global=False)
    assert env == check_alias

    # Check aliased env vars are passed including global hab env vars
    env = new_dict()
    cfg.update_environ(env, alias_name="as_str", include_global=True)
    check_freeze(env)
    assert env == check_global

    # Check aliased env vars are passed including global hab env vars
    env = new_dict()
    cfg.update_environ(env, alias_name="global", include_global=True)
    check = dict(check_global, **check_alias)
    # This env var is un-set by the alias
    del check["ALIASED_GLOBAL_D"]
    check_freeze(env)
    assert env == check


class TestCustomVariables:
    def test_distro(self, uncached_resolver):
        """Test that a distro processes valid custom variables correctly."""
        distro = uncached_resolver.distros["maya2024"].latest_version("maya2024")
        assert distro.name == "maya2024==2024.0"

        # Check that the custom variables are assigned
        check = {
            "maya_root_linux": "/usr/autodesk/maya2024/bin",
            "maya_root_windows": "C:/Program Files/Autodesk/Maya2024/bin",
        }
        assert distro.variables == check

        # Check that the aliases have not had their custom vars replaced
        alias = distro.aliases["windows"][0]
        assert alias[0] == "maya"
        assert alias[1]["cmd"] == "{maya_root_windows}/maya.exe"
        alias = distro.aliases["linux"][0]
        assert alias[0] == "maya"
        assert alias[1]["cmd"] == "{maya_root_linux}/maya"

        # Check that the custom variables are replaced when formatting
        formatted = distro.format_environment_value(distro.aliases)

        alias = formatted["windows"][0]
        assert alias[0] == "maya"
        assert alias[1]["cmd"] == f"{check['maya_root_windows']}/maya.exe"
        alias = formatted["linux"][0]
        assert alias[0] == "maya"
        assert alias[1]["cmd"] == f"{check['maya_root_linux']}/maya"

    @pytest.mark.parametrize("config_class", ("Config", "FlatConfig"))
    def test_config(self, uncached_resolver, config_class):
        """Test that a config processes valid custom variables correctly."""
        if config_class == "Config":
            cfg = uncached_resolver.closest_config("project_a")
            assert type(cfg) == Config
        else:
            cfg = uncached_resolver.resolve("project_a")
            assert type(cfg) == FlatConfig

        check = {
            "mount_linux": "/blur/g",
            "mount_windows": "G:",
        }
        assert cfg.variables == check

        cfg.environment
        env = cfg.frozen_data["environment"]
        # Check the mount_linux variable was replaced correctly
        assert env["linux"]["HOUDINI_OTLSCAN_PATH"] == [
            "/blur/g/project_a/cfg/hdas",
            "/blur/g/_shared/cfg/hdas",
            "&",
        ]
        assert env["linux"]["OCIO"] == ["/blur/g/project_a/cfg/ocio/v0001/config.ocio"]

        # Check the mount_windows variable was replaced correctly
        assert env["windows"]["HOUDINI_OTLSCAN_PATH"] == [
            "G:/project_a/cfg/hdas",
            "G:/_shared/cfg/hdas",
            "&",
        ]
        assert env["windows"]["OCIO"] == ["G:/project_a/cfg/ocio/v0001/config.ocio"]

    @pytest.mark.parametrize(
        "variables,invalid",
        (
            ({"relative_root": "Not Valid"}, "relative_root"),
            ({"relative_root": "Not Valid", ";": "Not Valid"}, ";, relative_root"),
            ({"valid": "Valid", ";": "Not Valid"}, ";"),
        ),
    )
    def test_reserved(self, uncached_resolver, variables, invalid, tmpdir):
        """Test that using a reserved variable raises an exception."""

        # Create a test distro file using the given reserved variables
        template = {
            "name": "maya2024",
            "version": "2024.99",
            "variables": variables,
        }
        distro_file = tmpdir / "maya2024" / ".hab.json"
        Path(distro_file).parent.mkdir()
        with distro_file.open("w") as fle:
            json.dump(template, fle, indent=4, cls=utils.HabJsonEncoder)

        # Add the test distro to hab's distro search. We don't need to call
        # `clear_caches` because distros haven't been resolved yet.
        uncached_resolver.distro_paths.append(Path(tmpdir))

        # When distros are resolved, an exception should be raised
        with pytest.raises(
            ReservedVariableNameError,
            match=rf"'{invalid}' are reserved variable name\(s\) for hab",
        ):
            uncached_resolver.distros
